;    awnetwork.pbi
;    Version 0.4 [2013/08/21]
;    Copyright (C) 2013 Ronny Krueger
;
;    This file is part of AWPB-Tools.
;
;    AWPB-Tools is free software: you can redistribute it and/or modify
;    it under the terms of the GNU Lesser General Public License as published by
;    the Free Software Foundation, either version 3 of the License, or
;    (at your option) any later version.
;
;    AWPB-Tools is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU Lesser General Public License for more details.
;
;    You should have received a copy of the GNU Lesser General Public License
;    along with AWPB-Tools.  If not, see <http://www.gnu.org/licenses/>.

CompilerIf Defined(Enable_OpenSSL, #PB_Constant)
  XIncludeFile "awopenssl.pbi"
CompilerEndIf

;{ Structures
Structure ____NW_NetworkConn
  ConnID.i
  ResultString.s
  ResultCode.i
  DataString.s
  TimeOut.i
  Retries.i
  Buffer.i
  Host.s
  Port.w
  User.s
  Pass.s
  EnableLog.i
  LogData.s
  CompilerIf Defined(Enable_OpenSSL, #PB_Constant)
  	TLSactive.i
  	sslHandle.i		; SSL*
  	sslContext.i	; SSL_CTX*
  	socket.i
  	lastErr.l
	CompilerEndIf
EndStructure
;}
;{ Constants
Enumeration
	#____AWPB_SSL_METHOD_V2
	#____AWPB_SSL_METHOD_V3
	#____AWPB_SSL_METHOD_V23
	#____AWPB_SSL_METHOD_TLSV1
EndEnumeration

#____AWPB_SSL_ERR_NONE				=  0
#____AWPB_SSL_UNKNOWN_METHOD	= -1
#____AWPB_SSL_CTX_NEW_FAIL		= -2
#____AWPB_SSL_SSL_NEW_FAIL		= -3
#____AWPB_SSL_SET_FD_FAIL			= -4
#____AWPB_SSL_CONNECT_FAIL		= -5
#____AWPB_SSL_WRITE           = -6

#____NW_BUF_SIZE  = 15000
#____NW_CRLF      = Chr(13) + Chr(10)
#____NW_CD_ERR    = 0
#____NW_CD_OK     = -1;}
;{ private functions
Procedure ____NW_AddLog(*nwconn.____NW_NetworkConn, text.s)
	If *nwconn\EnableLog
		*nwconn\LogData + text + #____NW_CRLF
	EndIf
EndProcedure
Procedure ____NW_AddLogErr(*nwconn.____NW_NetworkConn, text.s, errcode.l)
	If *nwconn\EnableLog
		*nwconn\LogData + text + " (Error Code: " + Str(errcode) + ")" + #____NW_CRLF
	EndIf
EndProcedure
Procedure.i ____NW_GetResultCode(*nwconn.____NW_NetworkConn)
  Protected code.i

  If LCase(Left(*nwconn\ResultString, 3)) = "+ok"
    code = #____NW_CD_OK
  ElseIf LCase(Left(*nwconn\ResultString, 4)) = "-err"
    code = #____NW_CD_ERR
  Else
    code = Val(Left(*nwconn\ResultString, 3))
  EndIf

  ProcedureReturn code
EndProcedure

CompilerIf Defined(Enable_OpenSSL, #PB_Constant)
	Procedure.l ____NW_SSL_Connect(*nwconn.____NW_NetworkConn, sslMethod.i)
		Protected res.l

		*nwconn\socket     = ConnectionID(*nwconn\ConnID)
  	*nwconn\TLSactive  = #False
  	*nwconn\sslHandle  = #Null
  	*nwconn\sslContext = #Null

		Select sslMethod
			Case #____AWPB_SSL_METHOD_V2:			*nwconn\sslContext = SSL_CTX_new(SSLv2_client_method())
			Case #____AWPB_SSL_METHOD_V3:			*nwconn\sslContext = SSL_CTX_new(SSLv3_client_method())
			Case #____AWPB_SSL_METHOD_V23:		*nwconn\sslContext = SSL_CTX_new(SSLv23_client_method())
			Case #____AWPB_SSL_METHOD_TLSV1:	*nwconn\sslContext = SSL_CTX_new(TLSv1_client_method())
			Default:													____NW_AddLog(*nwconn, "Unknown SSL Method requested")
																				ProcedureReturn #____AWPB_SSL_UNKNOWN_METHOD
		EndSelect

		If *nwconn\sslContext = #Null
			____NW_AddLog(*nwconn, "Unable to create SSL Context")
			ProcedureReturn #____AWPB_SSL_CTX_NEW_FAIL
		EndIf

		*nwconn\sslHandle = SSL_new(*nwconn\sslContext)
		If *nwconn\sslHandle = #Null
			____NW_AddLog(*nwconn, "Unable to obtain SSL Handle")
			ProcedureReturn #____AWPB_SSL_SSL_NEW_FAIL
		EndIf

		If Not SSL_set_fd(*nwconn\sslHandle, *nwconn\socket)
			____NW_AddLog(*nwconn, "Unable to connect socket to SSL")
			ProcedureReturn #____AWPB_SSL_SET_FD_FAIL
		EndIf

		res = SSL_connect(*nwconn\sslHandle)
		If res <> 1
			If res < 0
				res = SSL_get_error(*nwconn\sslHandle, res)
				____NW_AddLogErr(*nwconn, "Unable to connect with SSL", res)
				ProcedureReturn res
			EndIf
			ProcedureReturn #____AWPB_SSL_CONNECT_FAIL
		EndIf

		ProcedureReturn #____AWPB_SSL_ERR_NONE
	EndProcedure
	Procedure.l ____NW_SSL_Disconnect(*nwconn.____NW_NetworkConn)
		Protected res.l, erg.l

		erg = #____AWPB_SSL_ERR_NONE
		If *nwconn\sslHandle <> #Null
			res = SSL_shutdown(*nwconn\sslHandle)
			If res < 0
				erg = SSL_get_error(*nwconn\sslHandle, res)
				____NW_AddLogErr(*nwconn, "Unable to shutdown SSL", erg)
			EndIf
			SSL_free(*nwconn\sslHandle)
		EndIf

		If *nwconn\sslContext <> #Null
			SSL_CTX_free(*nwconn\sslContext)
		EndIf
  	*nwconn\TLSactive  = #False
  	*nwconn\sslHandle  = #Null
  	*nwconn\sslContext = #Null
  	*nwconn\socket     = #Null

  	ProcedureReturn erg
  EndProcedure
  Procedure.b ____NW_SendNetworkBin_SSL(*nwconn.____NW_NetworkConn, *buf, buflen.i)
  	Protected maxlen.l, sentlen.l, timeout.i, orglen.i, err.l

	  ; sendet binrdaten; dies geschieht in Blcken von maximal 50000 Byte

	  *nwconn\lastErr	= #____AWPB_SSL_ERR_NONE
	  orglen  				= buflen
	  timeout 				= *nwconn\TimeOut
	  While (buflen > 0) And (timeout > 0)
	    maxlen  = buflen
	    If maxlen > 50000 : maxlen = 50000 : EndIf

	    Repeat
		    sentlen = SSL_write(*nwConn\sslHandle, *buf, maxlen)
		    err = SSL_get_error(*nwConn\sslHandle, sentlen)
			Until (err <> #SSL_ERROR_WANT_WRITE) And err <> (#SSL_ERROR_WANT_READ)

			____NW_AddLog(*nwconn, "SSL written " + Str(sentlen) + " from " + Str(maxlen) + " bytes")

	    If err = #SSL_ERROR_ZERO_RETURN
				____NW_AddLogErr(*nwconn, "SSL_write", err)
	    	ProcedureReturn #False
	    ElseIf (err <> 0)
	    	*nwconn\lastErr = err
	    	____NW_AddLogErr(*nwconn, "SSL_write", err)
	    	ProcedureReturn #False
	    Else
	    	If sentlen > 0
		      buflen - sentlen
		      *buf   + sentlen
	  	  Else
	    	  Delay(50) : timeout - 50
	    	EndIf
	    EndIf
	  Wend
	  If buflen > 0
	  	*nwconn\lastErr = #____AWPB_SSL_WRITE
	  	____NW_AddLog(*nwconn, "SendNetworkBin_SSL not written all bytes")
	  	ProcedureReturn #False
	  EndIf
	  ProcedureReturn #True
  EndProcedure
	Procedure.s ____NW_ReceiveNetworkStr_SSL(*nwconn.____NW_NetworkConn)
	  Protected len.l, err.l

	  Repeat
	  	len = SSL_read(*nwconn\sslHandle, *nwconn\Buffer, #____NW_BUF_SIZE)
	  	err = SSL_get_error(*nwconn\sslHandle, len)
	  Until (err <> #SSL_ERROR_WANT_READ) And (err <> #SSL_ERROR_WANT_WRITE)
	  *nwconn\lastErr = err

	  ____NW_AddLog(*nwconn, "SSL read " + Str(len) + " bytes")

	  If len > 0
	  	____NW_AddLog(*nwconn, "SSL received data: >>" + PeekS(*nwconn\Buffer, len, #PB_Ascii) + "<<")
	  	ProcedureReturn PeekS(*nwconn\Buffer, len, #PB_Ascii)
	  EndIf

	  ____NW_AddLogErr(*nwconn, "SSL_read", *nwconn\lastErr)

	  ProcedureReturn #NULL$
	EndProcedure
	Procedure   ____NW_ReadResult_SSL(*nwconn.____NW_NetworkConn)
		Protected dt.s, done.i

    *nwconn\ResultString = ____NW_ReceiveNetworkStr_SSL(*nwconn)
    *nwconn\ResultCode   = ____NW_GetResultCode(*nwconn)
    *nwconn\DataString   = Mid(*nwconn\ResultString, FindString(*nwconn\ResultString, Chr(10), 1) + 1)
    If (*nwconn\lastErr = #SSL_ERROR_NONE) And (SSL_pending(*nwconn\sslHandle) < 1)
    	done = #True
    Else
    	done = #False
    EndIf
    While Not done
    	dt = ____NW_ReceiveNetworkStr_SSL(*nwconn)

    	If *nwconn\lastErr = #SSL_ERROR_ZERO_RETURN
	  		done = #True
	  	ElseIf *nwconn\lastErr = #SSL_ERROR_NONE
	  		If dt : *nwconn\DataString + dt : EndIf
	  		If SSL_pending(*nwconn\sslHandle) < 1
	  			done = #True
	  		EndIf
	  	EndIf
	  Wend
	EndProcedure
	Procedure   ____NW_ReadData_SSL(*nwconn.____NW_NetworkConn)
	  Protected dt.s, retry.i, done.i

	  dt   = ""
	  done = #False
    Repeat
    	If *nwconn\lastErr = #SSL_ERROR_ZERO_RETURN
	  		done = #True
	  	ElseIf *nwconn\lastErr = #SSL_ERROR_NONE
	  		If dt : *nwconn\DataString + dt : EndIf
			EndIf
			If (Right(*nwconn\DataString, 5) = #____NW_CRLF + "." + #____NW_CRLF) : done = #True : EndIf
    	If Not done : dt = ____NW_ReceiveNetworkStr_SSL(*nwconn) : EndIf
		Until done
	EndProcedure
CompilerEndIf

Procedure.b ____NW_SendNetworkBin_PLAIN(*nwconn.____NW_NetworkConn, *buf, buflen.i)
  Protected maxlen.l, sentlen.l, timeout.i, orglen.i, err.l

  ; sendet binrdaten; dies geschieht in Blcken von maximal 50000 Byte

  orglen  = buflen
  timeout = *nwconn\TimeOut
  While (buflen > 0) And (timeout > 0)
    maxlen  = buflen
    If maxlen > 50000 : maxlen = 50000 : EndIf
    sentlen = SendNetworkData(*nwconn\ConnID, *buf, maxlen)

    ____NW_AddLog(*nwconn, "connection written " + Str(sentlen) + " from " + Str(maxlen) + " bytes")

		If sentlen > 0
      buflen - sentlen
      *buf   + sentlen
    Else
      Delay(50) : timeout - 50
    EndIf
  Wend

  If buflen > 0
  	____NW_AddLog(*nwconn, "SendNetworkBin_PLAIN not written all bytes")
  	ProcedureReturn #False
  EndIf

  ProcedureReturn #True
EndProcedure
Procedure.b ____NW_SendNetworkBin(*nwconn.____NW_NetworkConn, *buf, buflen.i)
	Protected res.b

	CompilerIf Defined(Enable_OpenSSL, #PB_Constant)
		If *nwconn\TLSactive
			res = ____NW_SendNetworkBin_SSL(*nwconn, *buf, buflen)
		Else
			res = ____NW_SendNetworkBin_PLAIN(*nwconn, *buf, buflen)
		EndIf
	CompilerElse
		res = ____NW_SendNetworkBin_PLAIN(*nwconn, *buf, buflen)
	CompilerEndIf

	ProcedureReturn res
EndProcedure
Procedure.b ____NW_SendNetworkStr(*nwconn.____NW_NetworkConn, msg.s, crlf.b)
  Protected ok.b, l.i, crlfvalue.w

  ; sendet den bergebenen String
  ; im Unicode-Modus wird er dazu vorher in einen ASCII-String konvertiert
  ; anschlieend wird, falls crlf = #true, ein CRLF gesendet

  ok = #True
  l  = Len(msg)

  CompilerIf #PB_Compiler_Unicode
    Protected  a.i, b.i, c.i

    If l <= #____NW_BUF_SIZE
      PokeS(*nwconn\Buffer, msg, l, #PB_Ascii)
      ok = ____NW_SendNetworkBin(*nwconn, *nwconn\Buffer, l)
    Else
      c = Int(l / #____NW_BUF_SIZE)
      b = l % #____NW_BUF_SIZE

      For a = 0 To c - 1
        PokeS(*nwconn\Buffer, Mid(msg, (a * #____NW_BUF_SIZE) + 1, #____NW_BUF_SIZE), #PB_Ascii)
        ok = ____NW_SendNetworkBin(*nwconn, *nwconn\Buffer, #____NW_BUF_SIZE)
        If Not ok : Break : EndIf
      Next a

      If b > 0 And ok
        PokeS(*nwconn\Buffer, Mid(msg, (c * #____NW_BUF_SIZE) + 1, b), #PB_Ascii)
        ok = ____NW_SendNetworkBin(*nwconn, *nwconn\Buffer, b)
      EndIf
    EndIf
  CompilerElse
    ok = ____NW_SendNetworkBin(*nwconn, @msg, l)
  CompilerEndIf

  If crlf
    crlfvalue = $0a0d
    ____NW_SendNetworkBin(*nwconn, @crlfvalue, 2)
  EndIf

  ____NW_AddLog(*nwconn, "sent data >>" + msg + "<<")

  ProcedureReturn ok
EndProcedure
Procedure.s ____NW_ReceiveNetworkStr_PLAIN(*nwconn.____NW_NetworkConn)
  Protected len.l, result.s, err.l

  result = ""
  Repeat
		len = ReceiveNetworkData(*nwconn\ConnID, *nwconn\Buffer, #____NW_BUF_SIZE)
    If len
      result + PeekS(*nwconn\Buffer, len, #PB_Ascii)
    EndIf
  Until len < #____NW_BUF_SIZE

  ____NW_AddLog(*nwconn, "connection read " + Str(len) + " bytes")
  ____NW_AddLog(*nwconn, "connection received data: >>" + result + "<<")

	ProcedureReturn result
EndProcedure
Procedure.s ____NW_ReceiveNetworkStr(*nwconn.____NW_NetworkConn)
  Protected result.s

  result = ""
	CompilerIf Defined(Enable_OpenSSL, #PB_Constant)
		If *nwconn\TLSactive
			result = ____NW_ReceiveNetworkStr_SSL(*nwconn)
		Else
			result = ____NW_ReceiveNetworkStr_PLAIN(*nwconn)
		EndIf
	CompilerElse
		result = ____NW_ReceiveNetworkStr_PLAIN(*nwconn)
	CompilerEndIf

  ProcedureReturn result
EndProcedure
Procedure.i ____NW_NetworkClientEvent(*nwconn.____NW_NetworkConn)
	ProcedureReturn NetworkClientEvent(*nwconn\ConnID)
EndProcedure
Procedure.i ____NW_WaitNetworkClientEvent(*nwconn.____NW_NetworkConn)
  Protected Event.i, TimeOut.i

  TimeOut = *nwconn\TimeOut
  Repeat
  	Event = ____NW_NetworkClientEvent(*nwconn)
    If Not event
      Delay(50)
      TimeOut - 50
    EndIf
  Until Event Or TimeOut <= 0

  ProcedureReturn Event
EndProcedure
Procedure   ____NW_ReadResult_PLAIN(*nwconn.____NW_NetworkConn)
  Protected dt.s

  *nwconn\ResultString = ""
  *nwconn\DataString   = ""
  *nwconn\ResultCode   = -1

	If ____NW_WaitNetworkClientEvent(*nwconn) = #PB_NetworkEvent_Data
    *nwconn\ResultString = ____NW_ReceiveNetworkStr_PLAIN(*nwconn)
    *nwconn\ResultCode   = ____NW_GetResultCode(*nwconn)
    *nwconn\DataString   = Mid(*nwconn\ResultString, FindString(*nwconn\ResultString, Chr(10), 1) + 1)
    While ____NW_NetworkClientEvent(*nwconn) = #PB_NetworkEvent_Data
      dt = ____NW_ReceiveNetworkStr_PLAIN(*nwconn)
      If dt : *nwconn\DataString + dt : EndIf
    Wend
  EndIf
EndProcedure
Procedure.i ____NW_ReadResult(*nwconn.____NW_NetworkConn)
	CompilerIf Defined(Enable_OpenSSL, #PB_Constant)
		If *nwconn\TLSactive
			____NW_ReadResult_SSL(*nwconn)
		Else
			____NW_ReadResult_PLAIN(*nwconn)
		EndIf
	CompilerElse
		____NW_ReadResult_PLAIN(*nwconn)
	CompilerEndIf

  ProcedureReturn *nwconn\ResultCode
EndProcedure
Procedure   ____NW_ReadData_PLAIN(*nwconn.____NW_NetworkConn)
  Protected dt.s, retry.i

  ;  retry = *nwconn\Retries
  While (Right(*nwconn\DataString, 5) <> #____NW_CRLF + "." + #____NW_CRLF) ; And (retry > 0)
    If ____NW_NetworkClientEvent(*nwconn) = #PB_NetworkEvent_Data
      dt = ____NW_ReceiveNetworkStr(*nwconn)
      If dt : *nwconn\DataString + dt : EndIf
    Else
      Delay(50)
;     retry - 1
		EndIf
	Wend
EndProcedure
Procedure   ____NW_ReadData(*nwconn.____NW_NetworkConn)
	CompilerIf Defined(Enable_OpenSSL, #PB_Constant)
		If *nwconn\TLSactive
			____NW_ReadData_SSL(*nwconn)
		Else
			____NW_ReadData_PLAIN(*nwconn)
		EndIf
	CompilerElse
		____NW_ReadData_PLAIN(*nwconn)
	CompilerEndIf
EndProcedure
Procedure.b ____NW_Execute(*nwconn.____NW_NetworkConn, cmd.s, retries.i = 1)
  Protected res.b

  Repeat
  	Delay(100)
  	res = ____NW_SendNetworkStr(*nwconn, cmd, #True)
    Delay(100)
    ____NW_ReadResult(*nwconn)
    If Not res
      retries - 1
    EndIf
  Until res Or (retries <= 0)

  ProcedureReturn res
EndProcedure
;}

